#ifndef _467gpu_h
#include <sys/types.h>
#include <stdbool.h>

// Opcode list
// You may want to alter the list of allowed opcodes to better suit your purposes.
#define LDGPMEM 0x01
#define STGPMEM 0x02
#define MUL     0x03
#define ADD     0x04
#define SRL     0x05
#define SLL     0x06
#define AND     0x07
#define NOT     0x08
#define XOR     0x09
#define OR      0x0a
#define NAND    0x0b
#define LI      0x0c
#define SETLT   0x0d
#define STOREQ  0x0e
#define END     0x0f
#define SUB     0x1f

// This is the list of parseable instructions.  You may want to
// alter this to add/remove instructions and allowable forms.
#define REG1        0x01
#define REG2        0x02
#define REG3        0x04
#define CONSTANT    0x08
#define MEMORY      0x10
#define PREDICATE   0x20

struct  instruction {
    char    *mnemonic;
    u_int32_t   opcode;
    u_int32_t   operand0; 
    u_int32_t   operand1; 
    u_int32_t   operand2; 
};

extern struct instruction instructions[];

#define MAX_CONSTANT    0xffff

// Encoding
// You may want to alter this encoding to suit your purposes.
//
//  [PREDICATE_2][OP3_1_TYPE][OPCODE_5]
//  [REG1_4][REG2_4]
//  [REG3_4][PADDING_12] OR [CONSTANT_16]
#define PREDICATE_SHIFT 30
#define PREDICATE_MASK  0x3
#define OP3_TYPE_SHIFT  29
#define OP3_TYPE_MASK   0x1
#define OPCODE_SHIFT    24
#define OPCODE_MASK     0x1f
#define REG1_SHIFT      20
#define REG1_MASK       0xf
#define REG2_SHIFT      16
#define REG2_MASK       0xf
#define REG3_SHIFT      12
#define REG3_MASK       0xf
#define CONSTANT_SHIFT  0
#define CONSTANT_MASK   0xffff

u_int32_t    encode_instruction(
    u_int32_t   opcode,
    u_int32_t   predicate,
    bool        use_constant,
    u_int32_t   constant,
    u_int32_t   reg1,
    u_int32_t   reg2,
    u_int32_t   reg3);

void decode_instruction(u_int32_t res,
    u_int32_t   *opcode,
    u_int32_t   *predicate,
    bool        *use_constant,
    u_int32_t   *constant,
    u_int32_t   *reg1,
    u_int32_t   *reg2,
    u_int32_t   *reg3);
        
bool disasm_instruction(u_int32_t   res, char *s);

#define MAX_LINE_LENGTH (16*1024)

struct debug_symbol {
    int     line_number;
    int     line_length;
    char    line[MAX_LINE_LENGTH];
};

#define NUMBER_OF_REGISTERS     16
#define NUMBER_OF_PREDICATES    4
#define CODE_SIZE               (16*1024)
#define GPMEM_SIZE              (16*1024*2)
#define QUEUE_IO                8   // number of words in a queue IO operation
#define MSR_REGISTER            15
struct context {
    u_int32_t   regs[NUMBER_OF_REGISTERS];
    union {
        u_int32_t   msr;
        struct  {
            u_int32_t   predicate0:1;    // hard wired to true
            u_int32_t   predicate1:1;
            u_int32_t   predicate2:1;
            u_int32_t   predicate3:1;
            u_int32_t   queue_src:2;
            u_int32_t   padding:2;
            u_int32_t   pid:8;
            u_int32_t   pc:16;
        };
    };
};

enum single_step_result_t {
    at_end = 0,
    normal_execution = 1,
    skipped_execution = 2,
    queue_write = 3,
    invalid_predicate = -1,
    invalid_address = -2,
};

enum single_step_result_t single_step(struct context *ctx, u_int32_t *code, u_int32_t *gpmem, u_int32_t *queue_data, int *queue);

#endif
